import time
import inspect
import logging

from tango import DevState, AttrQuality
from tango.server import Device, attribute, command, device_property

from sockio.sio import TCP

from .cryocon import CryoCon


def create_connection(address, connection_timeout=0.2, timeout=0.2):
    if address.startswith('tcp://'):
        address = address[6:]
        pars = address.split(':')
        host = pars[0]
        port = int(pars[1]) if len(pars) else 5000
        conn = TCP(host, port, connection_timeout=connection_timeout, timeout=timeout)
        return conn
    else:
        raise NotImplementedError(
            'address {!r} not supported'.format(address))


def attr(**kwargs):
    name = kwargs['name'].lower()
    func = ATTR_MAP[name]
    dtype = kwargs.setdefault('dtype', float)

    def get(self):
        value = self.last_values[name]
        if isinstance(value, Exception):
            raise value
        value = self.last_values[name]
        if value is None:
            value = float('nan') if dtype == float else ''
            return value, time.time(), AttrQuality.ATTR_INVALID
        return value

    attr = attribute(get, **kwargs)

    sig = inspect.signature(func)
    if len(sig.parameters) > 1:
        @attr.setter
        def fset(self, value):
            func(self.cryocon, value)
        fset.__name__ = 'write_' + name
        kwargs['fset'] = fset

    return attr


ATTR_MAP = {
    'idn': lambda cryo: cryo.idn(),
    'control': lambda cryo, v=None: cryo.control(v),
    'channela': lambda cryo: cryo['A'].temperature(),
    'channelb': lambda cryo: cryo['B'].temperature(),
    'channelc': lambda cryo: cryo['C'].temperature(),
    'channeld': lambda cryo: cryo['D'].temperature(),
    'loop1output': lambda cryo, v=None: cryo[1].output_power(v),
    'loop2output': lambda cryo, v=None: cryo[2].output_power(v),
    'loop3output': lambda cryo, v=None: cryo[3].output_power(v),
    'loop4output': lambda cryo, v=None: cryo[4].output_power(v),
    'loop1range': lambda cryo, v=None: cryo[1].range(v),
    'loop1rate': lambda cryo, v=None: cryo[1].rate(v),
    'loop2rate': lambda cryo, v=None: cryo[2].rate(v),
    'loop3rate': lambda cryo, v=None: cryo[3].rate(v),
    'loop4rate': lambda cryo, v=None: cryo[4].rate(v),
    'loop1type': lambda cryo, v=None: cryo[1].type(v),
    'loop2type': lambda cryo, v=None: cryo[2].type(v),
    'loop3type': lambda cryo, v=None: cryo[3].type(v),
    'loop4type': lambda cryo, v=None: cryo[4].type(v),
    'loop1setpoint': lambda cryo, v=None: cryo[1].set_point(v),
    'loop2setpoint': lambda cryo, v=None: cryo[2].set_point(v),
    'loop3setpoint': lambda cryo, v=None: cryo[3].set_point(v),
    'loop4setpoint': lambda cryo, v=None: cryo[4].set_point(v),
    'loop1pgain': lambda cryo: cryo[1].p_gain(),
    'loop2pgain': lambda cryo: cryo[2].p_gain(),
    'loop3pgain': lambda cryo: cryo[3].p_gain(),
    'loop4pgain': lambda cryo: cryo[4].p_gain(),
    'loop1igain': lambda cryo: cryo[1].i_gain(),
    'loop2igain': lambda cryo: cryo[2].i_gain(),
    'loop3igain': lambda cryo: cryo[3].i_gain(),
    'loop4igain': lambda cryo: cryo[4].i_gain(),
    'loop1dgain': lambda cryo: cryo[1].d_gain(),
    'loop2dgain': lambda cryo: cryo[2].d_gain(),
    'loop3dgain': lambda cryo: cryo[3].d_gain(),
    'loop4dgain': lambda cryo: cryo[4].d_gain(),
}


class CryoConTempController(Device):

    address = device_property(str)
    UsedChannels = device_property([str], default_value='ABCD')
    UsedLoops = device_property([int], default_value=[1, 2, 3, 4])
    ReadValidityPeriod = device_property(float, default_value=0.1)
    AutoLockFrontPanel = device_property(bool, default_value=False)

    def init_device(self):
        super().init_device()
        self.last_read_time = 0
        self._temperatures = None
        channels = ''.join(self.UsedChannels)
        loops = self.UsedLoops

        conn = create_connection(self.address)
        self.cryocon = CryoCon(conn, channels=channels, loops=loops)
        self.last_values = {}
        self.last_state_ts = 0

    def delete_device(self):
        super().delete_device()
        try:
            self.cryocon._conn.close()
        except Exception:
            logging.exception('Error closing cryocon')

    def read_attr_hardware(self, indexes):
        multi_attr = self.get_device_attr()
        names = ['control']
        with self.cryocon as group:
            self.cryocon.control()
            for index in sorted(indexes):
                attr = multi_attr.get_attr_by_ind(index)
                attr_name = attr.get_name().lower()
                func = ATTR_MAP[attr_name]
                func(self.cryocon)
                names.append(attr_name)
        values = group.replies
        self.last_values = dict(zip(names, values))
        self._update_state_status(self.last_values['control'])

    def _update_state_status(self, value=None):
        if value is None:
            ts = time.time()
            if ts < (self.last_state_ts + 1):
                return self.get_state(), self.get_status()
            try:
                value = self.cryocon.control()
            except Exception as error:
                value = error
        ts = time.time()
        if isinstance(value, Exception):
            state, status = DevState.FAULT, 'Error: {!r}'.format(value)
        else:
            state = DevState.ON if value else DevState.OFF
            status = 'Control is {}'.format('On' if value else 'Off')
        self.set_state(state)
        self.set_status(status)
        self.last_state_ts = ts
        self.__local_status = status  # prevent deallocation by keeping reference
        return state, status

    def dev_state(self):
        state, status = self._update_state_status()
        return state

    def dev_status(self):
        state, status = self._update_state_status()
        return status

    idn = attr(name='idn', dtype=str)
    channelA = attr(name='channelA')
    channelB = attr(name='channelB')
    channelC = attr(name='channelC')
    channelD = attr(name='channelD')
    loop1output = attr(name='loop1output')
    loop2output = attr(name='loop2output')
    loop3output = attr(name='loop3output')
    loop4output = attr(name='loop4output')
    loop1range = attr(name='loop1range')
    loop1rate = attr(name='loop1rate')
    loop2rate = attr(name='loop2rate')
    loop3rate = attr(name='loop3rate')
    loop4rate = attr(name='loop4rate')
    loop1type = attr(name='loop1type', dtype=str)
    loop2type = attr(name='loop2type', dtype=str)
    loop3type = attr(name='loop3type', dtype=str)
    loop4type = attr(name='loop4type', dtype=str)
    loop1setpoint = attr(name='loop1setpoint')
    loop2setpoint = attr(name='loop2setpoint')
    loop3setpoint = attr(name='loop3setpoint')
    loop4setpoint = attr(name='loop4setpoint')

    @command
    def on(self):
        self.cryocon.control(True)

    @command
    def off(self):
        self.cryocon.control(False)

    @command(dtype_in=str, dtype_out=str)
    def run(self, cmd):
        return self.cryocon._ask(cmd)

    @command(dtype_in=[str])
    def setchannelunit(self, unit):
        raise NotImplementedError


def main():
    import logging
    fmt = '%(asctime)s %(levelname)s %(threadName)s %(message)s'
    logging.basicConfig(level=logging.WARNING, format=fmt)
    CryoConTempController.run_server()


if __name__ == '__main__':
    main()



import time
import asyncio
import logging
import datetime
import functools
import contextlib


OUT_OF_RANGE = '_______'
OUT_OF_LIMIT = '.......'
NA = 'N/A'
DISABLED = ''
UNITS = ('K', 'C', 'F', 'S')
NACK = 'NACK'

# Delta read back tolerance
DELTA_RB = 0.0000001  # 6 decimals precision
# setpoints seem to have some kind of preset values when not in K or S units
DELTA_RB_SETPT = 0.0001

TYPES = ['OFF', 'PID', 'MAN', 'TABLE', 'RAMPP', 'RAMPT']
RANGES = ['HI', 'MID', 'LOW']


def to_int(text):
    if text in (OUT_OF_LIMIT, OUT_OF_RANGE, NA):
        return None
    return int(text)


def to_float(text):
    if text in (OUT_OF_LIMIT, OUT_OF_RANGE, NA):
        return None
    return float(text)


def to_float_unit(text):
    return to_float(text[:-1])


def to_on_off(text):
    return text.upper() == 'ON'


def from_on_off(b):
    if b in {True, 'on', 'ON'}:
        return 'ON'
    elif b in {False, 'off', 'OFF'}:
        return 'OFF'
    raise ValueError('Invalid ON/OFF value {!r}'.format(b))


def to_name(text):
    return text.strip('"')


def from_name(text):
    return '"{}"'.format(text)


def to_date(text):
    month, day, year = [int(i) for i in text.strip('"').split('/')]
    return datetime.date(year, month, day)


def from_date(date):
    if isinstance(date, str):
        if not date.startswith('"'):
            date = '"{}"'.format(date)
        return date
    return date.strftime('"%m/%d/%Y"')


def to_time(text):
    hh, mm, ss = [int(i) for i in text.strip('"').split(':')]
    return datetime.time(hh, mm, ss)


def from_time(time):
    if isinstance(time, str):
        if not time.startswith('"'):
            time = '"{}"'.format(time)
        return time
    return time.strftime('"%H:%M:%S"')


def handle_reply(reply):
    if reply is None:
        return
    reply = reply.decode().strip()
    if reply is NACK:
        raise CryoConError('Command not acknowledged')
    return reply


class CryoConError(Exception):
    pass


def sub_member(prefix, name, fget=lambda x: x, fset=None):
    assert not (fget is None and fset is None)
    cmd = ':{} {{}}:{}'.format(prefix.upper(), name.upper())

    def get_set(obj, value=None):
        command = cmd.format(obj.id)
        if value is None:
            if fget is None:
                raise ValueError('{} is not readable'.format(command))
            command += '?'
        elif fset is None:
            raise ValueError('{} is not writable'.format(command))
        else:
            set_command = '{} {}'.format(command, fset(value))
            if fget is None:
                return obj.ctrl._command(set_command)
            command = '{};{}?'.format(set_command, command)
        return obj.ctrl._query(command, fget)

    return get_set


channel_member = functools.partial(sub_member, 'INPUT')
loop_member = functools.partial(sub_member, 'LOOP')


class Channel:

    name = channel_member('nam', to_name, from_name)
    temperature = channel_member('temp', to_float)
    unit = channel_member('unit')
    minimum = channel_member('min', to_float)
    maximum = channel_member('max', to_float)
    variance = channel_member('vari', to_float)
    slope = channel_member('slop', to_float)
    offset = channel_member('offs', to_float)
    alarm = channel_member('alar')

    def __init__(self, channel, ctrl):
        self.id = channel
        self.ctrl = ctrl

    def clear_alarm(self):
        self.ctrl._command(':INPUT A:ALAR:CLE')


class Loop:

    source = loop_member('source', fset=str)
    set_point = loop_member('setpt', to_float_unit)
    error = loop_member('err')
    type = loop_member('typ', fset=str)
    range = loop_member('rang', fset=str)
    ramp = loop_member('ramp', to_on_off)
    rate = loop_member('rate', to_float, str)
    proportional_gain = loop_member('pga', to_float, str)
    integrator_gain = loop_member('iga', to_float, str)
    differentiator_gain = loop_member('dga', to_float, str)
    manual_output_power = loop_member('pman', to_float, str)  # percentage
    output_power = loop_member('outp', to_float)  # percentage
    load = loop_member('load', to_int, str)
    max_output_power = loop_member('maxp', to_float, str)  # percentage
    max_set_point = loop_member('maxs', to_float_unit)
    output_voltage = loop_member('vsen', to_float_unit)  # in V
    output_current = loop_member('isen', to_float_unit)  # in A
    output_load_resistance = loop_member('lsen', to_float, None)
    temperature = loop_member('htrh', to_float_unit, None)  # in degC
    autotune_status = loop_member('aut:stat', str, None)

    def __init__(self, nb, ctrl):
        self.id = nb
        self.ctrl = ctrl

    def _query(self, cmd, func=lambda x: x):
        cmd = ':LOOP {}:{}?'.format(self.id, cmd)
        return self.ctrl._query(cmd, func)

    def _command(self, cmd, value):
        cmd = ':LOOP {}:{} {}'.format(self.id, cmd, value)
        self.ctrl._command(cmd)


def member(name, fget=lambda x: x, fset=None):
    assert not (fget is None and fset is None)
    cmd = ':{}'.format(name.upper())

    def get_set(obj, value=None):
        command = cmd
        if value is None:
            if fget is None:
                raise ValueError('{} is not readable'.format(command))
            command += '?'
        elif fset is None:
            raise ValueError('{} is not writable'.format(command))
        else:
            set_command = '{} {}'.format(command, fset(value))
            if fget is None:
                return obj._command(set_command)
            command = '{};{}?'.format(set_command, command)
        return obj._query(command, fget)

    return get_set


class CryoCon:

    comm_error_retry_period = 3

    class Group:

        def __init__(self, ctrl):
            self.ctrl = ctrl
            self.cmds = ['']
            self.funcs = []

        def append(self, cmd, func):
            cmds = self.cmds[-1]
            # maximum of 255 characters per command
            if len(cmds) + len(cmd) > 250:
                cmds = ''
                self.cmds.append(cmds)
            cmds += cmd
            self.cmds[-1] = cmds
            self.funcs.append(func)

        def _store(self, replies):
            replies = ''.join(replies)
            replies = (msg.strip() for msg in replies.split(';'))
            replies = [func(text) for func, text in zip(self.funcs, replies)]
            self.replies = replies

        async def _async_store(self, replies):
            self._store([await reply for reply in replies])

        def query(self):
            replies = [self.ctrl._ask(request) for request in self.cmds]
            is_async = replies and asyncio.iscoroutine(replies[0])
            store = self._async_store if is_async else self._store
            return store(replies)

    def __init__(self, conn, channels='ABCD', loops=(1, 2, 3, 4)):
        self._conn = conn
        self._is_async = asyncio.iscoroutinefunction(conn.write_readline)
        self._log = logging.getLogger("CryoCon({}:{})".format(conn.host, conn.port))
        self._last_comm_error = None, 0  # (error, timestamp)
        self.channels = {channel: Channel(channel, self) for channel in channels}
        self.loops = {loop: Loop(loop, self) for loop in loops}
        self.group = None

    def __getitem__(self, key):
        try:
            return self.channels[key]
        except KeyError:
            return self.loops[key]

    def __enter__(self):
        self.group = self.Group(self)
        return self.group

    def __exit__(self, exc_type, exc_value, traceback):
        group = self.group
        self.group = None
        group.query()

    async def __aenter__(self):
        self.group = self.Group(self)
        return self.group

    async def __aexit__(self, exc_type, exc_value, traceback):
        group = self.group
        self.group = None
        await group.query()

    @contextlib.contextmanager
    def _guard_io(self):
        self._last_comm_error = None, 0
        try:
            yield
        except OSError as comm_error:
            self._last_comm_error = comm_error, time.time()
            raise

    async def _async_io(self, func, request):
        self._log.debug("REQ: %r", request)
        with self._guard_io():
            reply = handle_reply(await func(request))
        self._log.debug("REP: %r", reply)
        return reply

    def _sync_io(self, func, request):
        self._log.debug("REQ: %r", request)
        with self._guard_io():
            reply = handle_reply(func(request))
        self._log.debug("REP: %r", reply)
        return reply

    def _ask(self, cmd):
        now = time.time()
        last_err, last_ts = self._last_comm_error
        if now < (last_ts + self.comm_error_retry_period):
            raise last_err
        query = '?;' in cmd
        raw_cmd = cmd.encode() + b'\n'
        io = self._conn.write_readline if query else self._conn.write
        handle = self._async_io if asyncio.iscoroutinefunction(io) else self._sync_io
        return handle(io, raw_cmd)

    def _query(self, cmd, func=lambda x: x):
        cmd = cmd + ';'
        if self.group is None:
            reply = self._ask(cmd)
            if asyncio.iscoroutine(reply):
                async def async_func(reply):
                    return func(await reply)
                reply = async_func(reply)
            else:
                reply = func(reply)
            return reply
        else:
            self.group.append(cmd, func)

    def _command(self, cmd):
        return self._ask(cmd)

    idn = member('*IDN')
    name = member('SYSTEM:NAME', to_name, from_name)
    hw_revision = member('SYSTEM:HWR')
    fw_revision = member('SYSTEM:FWR')
    lockout = member('SYSTEM:LOCKOUT', to_on_off, from_on_off)
    led = member('SYSTEM:REMLED', to_on_off, from_on_off)
    display_filter_time = member('SYSTEM:DISTC', to_float, str)
    date = member('SYSTEM:DATE', to_date, from_date)
    time = member('SYSTEM:TIME', to_time, from_time)

    def control(self, value=None):
        cmd = ':CONTROL?'
        if value is not None:
            set_cmd = ':CONTROL' if value in (True, 'on', 'ON') else ':STOP'
            cmd = '{};{}'.format(set_cmd, cmd)
        return self._query(cmd, to_on_off)

    def __repr__(self):
        if self._is_async:
            data = '(asynchonous)'
        else:
            items = 'idn', 'name', 'control', 'hw_revision', 'fw_revision'
            try:
                with self as group:
                    for item in items:
                        getattr(self, item)()
                data = '\n'.join('{}: {}'.format(key.upper(), value)
                                 for key, value in zip(items, group.replies))
            except OSError:
                data = '(disconnected)'

        return 'CrycoCon({}:{})\n{}'.format(self._conn.host, self._conn.port, data)



# -*- coding: utf-8 -*-
#
# This file is part of the instrument simulator project
#
# Copyright (c) 2019 Tiago Coutinho
# Distributed under the MIT. See LICENSE for more info.

"""
.. code-block:: yaml

    devices:
    - class: CryoCon
      package: cryocon.simulator
      transports:
      - type: tcp
        url: :5000

A simple *nc* client can be used to connect to the instrument:

    $ nc 0 5000
    *IDN?
    Cryo-con,24C,204683,1.01A

Complex configuration with default values on simulator startup:

- class: CryoCon
  package: cryocon.simulator
  transports:
  - type: tcp
    url: :5001
  channels:
  - id: A
    unit: K
  - id: B
    unit: K
  loops:
  - id: 1
    source: A
    type: MAN
  distc: 4
  lockout: OFF
  remled: ON
  control: OFF
"""

import time
import random

import scpi
import gevent

from sinstruments.simulator import BaseDevice


DEFAULT_CHANNEL = {
    'unit': 'K',
    'alarm': '--',
    'alarm_highest': '105',
    'alarm_lowest': '5',
    'alarm_deadband': '0.1',
    'alarm_high_enabled': 'NO',
    'alarm_low_enabled': 'NO',
    'alarm_latch_enabled': 'NO',
    'alarm_audio': 'NO',
    'minimum': '3.4',
    'maximum': '304.12',
    'variance': '0.1',
    'slope': '0.5',
    'offset': '5.5',
}


def Channel(**data):
    id = data['id']
    data.setdefault('name', 'Channel' + id)
    channel = dict(DEFAULT_CHANNEL, **data)
    return channel


TEMPS = ['11.456', '12.456', '13.456', '14.456', '.......']

DEFAULT_LOOP = {
    'source': 'A',
    'type': 'MAN',
    'output power': '40.3',
    'setpoint': '0.0',
    'rate': '10.0',
    'range': 'MID'
}


def Loop(**data):
    return dict(DEFAULT_LOOP, **data)


DEFAULT = {
    '*idn': 'Cryo-con,24C,204683,1.01A',
    'name': 'Cryocon simulator',
    'lockout': 'OFF',
    'distc': 1,
    'remled': 'OFF',
    'control': 'OFF',
    'hardware_revision': '12A4FE',
    'firmware_revision': '78C90A',
    'channels': {name: Channel(id=name) for name in 'ABCD'},
    'loops': {str(loop): Loop(id=loop) for loop in range(2)}
}


class CryoCon(BaseDevice):

    MIN_TIME = 0.1

    def __init__(self, name, **opts):
        kwargs = {}
        if 'newline' in opts:
            kwargs['newline'] = opts.pop('newline')
        self._config = dict(DEFAULT, **opts)
        self._config['channels'] = {channel['id'].upper(): Channel(**channel)
                                    for channel in self._config['channels'].values()}
        self._config['loops'] = {str(loop['id']): Loop(**loop)
                                 for loop in self._config['loops'].values()}
        super().__init__(name, **kwargs)
        self._last_request = 0
        self._cmds = scpi.Commands({
            '*IDN': scpi.Cmd(get=lambda req: self._config['*idn']),
            'SYSTem:LOCKout': scpi.Cmd(get=self.lockout, set=self.lockout),
            'SYSTem:REMLed': scpi.Cmd(get=self.remled, set=self.remled),
            'SYSTem:NAMe': scpi.Cmd(get=self.sys_name, set=self.sys_name),
            'SYSTem:DATe': scpi.Cmd(get=self.sys_date, set=self.sys_date),
            'SYSTem:TIMe': scpi.Cmd(get=self.sys_time, set=self.sys_time),
            'SYSTem:HWRev': scpi.Cmd(get=self.hw_revision),
            'SYSTem:FWRev': scpi.Cmd(get=self.fw_revision),
            'CONTrol': scpi.Cmd(get=self.control, set=self.control),
            'STOP': scpi.Cmd(set=self.stop),
            'INPut': scpi.Cmd(get=self.get_input, set=self.set_input),
            'LOOP': scpi.Cmd(get=self.get_loop, set=self.set_loop),
        })

    def handle_line(self, line):
        self._log.debug('request %r', line)
        curr_time = time.time()
        dt = self.MIN_TIME - (curr_time - self._last_request)
        self._last_request = curr_time
        if dt > 0:
            self._log.debug('too short requests. waiting %.3f ms', dt*1000)
            gevent.sleep(dt)
        line = line.decode()
        requests = scpi.split_line(line)
        results = (self.handle_request(request) for request in requests)
        results = (result for result in results if result is not None)
        reply = ';'.join(results).encode()
        if reply:
            reply += b'\n'
            self._log.debug('reply %r', reply)
            return reply

    def handle_request(self, request):
        cmd = self._cmds.get(request.name)
        if cmd is None:
            return 'NACK'
        if request.query:
            getter = cmd.get('get')
            if getter is None:
                return 'NACK'
            return cmd['get'](request)
        else:
            setter = cmd.get('set')
            if setter is None:
                return 'NACK'
            return cmd['set'](request)

    def lockout(self, request):
        if request.query:
            return 'ON' if self._config['lockout'] in ('ON', True) else 'OFF'
        args = request.args.upper()
        if args in ('ON', 'OFF'):
            self._config['lockout'] = args

    def remled(self, request):
        if request.query:
            return 'ON' if self._config['remled'] in ('ON', True) else 'OFF'
        args = request.args.upper()
        if args in ('ON', 'OFF'):
            self._config['remled'] = args

    def control(self, request):
        if request.query:
            return 'ON' if self._config['control'] in ('ON', True) else 'OFF'
        self._config['control'] = 'ON'

    def stop(self, request):
        self._config['control'] = 'OFF'

    def sys_name(self, request):
        if request.query:
            return self._config['name']
        self._config['name'] = request.args.replace('"', '')

    def sys_date(self, request):
        if request.query:
            return time.strftime('"%m/%d/%Y"')
        # cannot change machine date!

    def sys_time(self, request):
        if request.query:
            return time.strftime('"%H:%M:%S"')
        # cannot change machine time!

    def hw_revision(self, request):
        return self._config['hardware_revision']

    def fw_revision(self, request):
        return self._config['firmware_revision']

    def get_input(self, request):
        if ':' in request.args:
            channels, variable = request.args.split(':', 1)
            variable = variable.upper()
        else:
            channels, variable = request.args, 'TEMP'
        channels = [ch.upper() for ch in channels.split(',')]
        if variable.startswith('TEMP'):
            values = [random.choice(TEMPS) for channel in channels]
            return ';'.join(values)
        elif variable.startswith('UNIT'):
            ch = self._config['channels']
            values = [ch[channel]['unit'] for channel in channels]
            return ';'.join(values)
        elif variable.startswith('NAM'):
            ch = self._config['channels']
            values = [ch[channel]['name'] for channel in channels]
            return ';'.join(values)
        elif variable.startswith('SLOP'):
            ch = self._config['channels']
            values = [ch[channel]['slope'] for channel in channels]
            return ';'.join(values)
        elif variable.startswith('VARI'):
            ch = self._config['channels']
            values = [ch[channel]['variance'] for channel in channels]
            return ';'.join(values)
        elif variable.startswith('MIN'):
            ch = self._config['channels']
            values = [ch[channel]['minimum'] for channel in channels]
            return ';'.join(values)
        elif variable.startswith('MAX'):
            ch = self._config['channels']
            values = [ch[channel]['maximum'] for channel in channels]
            return ';'.join(values)
        elif variable.startswith('ALAR'):
            ch = self._config['channels']
            values = [ch[channel]['alarm'] for channel in channels]
            return ';'.join(values)
        else:
            return 'NACK'

    def set_input(self, request):
        arg, value = request.args.split(' ', 1)
        channel, variable = arg.split(':', 1)
        variable = variable.upper()
        channels = self._config['channels']
        channel = channels[channel.upper()]
        if variable.startswith('UNIT'):
            channel['unit'] = value
        elif variable.startswith('NAM'):
            channel['name'] = value
        else:
            return 'NACK'

    def get_loop(self, request):
        channel, variable = request.args.split(':', 1)
        variable = variable.upper()
        loop = self._config['loops']
        channel = loop[channel]
        if variable.startswith('SOUR'):
            return channel['source']
        elif variable.startswith('SETP'):
            return channel['setpoint']
        elif variable.startswith('TYP'):
            return channel['type']
        elif variable.startswith('OUTP'):
            return channel['output power']
        elif variable.startswith('RAT'):
            return channel['rate']
        elif variable.startswith('RANG'):
            return channel['range']
        else:
            return 'NACK'

    def set_loop(self, request):
        arg, value = request.args.split(' ', 1)
        channel, variable = arg.split(':', 1)
        variable = variable.upper()
        loop = self._config['loops']
        channel = loop[channel]
        if variable.startswith('SOUR'):
            channel['source'] = value
        elif variable.startswith('SETP'):
            channel['setpoint'] = value
        elif variable.startswith('TYP'):
            channel['type'] = value
        elif variable.startswith('OUTP'):
            channel['output power'] = value
        elif variable.startswith('RAT'):
            channel['rate'] = value
        elif variable.startswith('RANG'):
            channel['range'] = value
        else:
            return 'NACK'
